use std::{collections::HashMap, path::Path};

use camino::Utf8PathBuf;
use wiremock::{
    matchers::{method, path},
    Mock, ResponseTemplate,
};

mod common;

#[tokio::test]
async fn test_ingest() {
    common::log_init();
    let (mock, solr) = common::solr_server().await;
    Mock::given(method("POST"))
        .and(path("/solr/files/update/json/docs"))
        .respond_with(ResponseTemplate::new(200))
        .mount(&mock)
        .await;
    Mock::given(method("POST"))
        .and(path("/solr/latest/update/json/docs"))
        .respond_with(ResponseTemplate::new(200))
        .mount(&mock)
        .await;

    let drs_config = common::test_config();
    let data_dir: Option<&Path> = None;

    let result = freva::drs::ingest(&solr, &drs_config, &data_dir, 10, None, None).await;
    assert!(result.is_ok());
    let report = result.unwrap();
    let expected_results: HashMap<&str, (u32, u32)> =
        HashMap::from_iter([("observations", (24, 0)), ("cmip5_name", (10, 0))].into_iter());

    assert_eq!(
        expected_results.len(),
        report.datasets.len(),
        "report did not have expected number of dataset reports"
    );
    for dataset_report in report.datasets.iter() {
        let name = dataset_report.dataset.name();
        let (sent, skipped) = expected_results.get(name).expect("unexpected dataset name");
        // can't assert_eq the whole struct because it includes a duration which is non-deterministic
        assert_eq!(
            *sent, dataset_report.sent,
            "unexpected sent number for dataset {}",
            name
        );
        assert_eq!(
            *skipped, dataset_report.skipped,
            "unexpected skipped number for dataset {}",
            name
        );
    }
}

#[tokio::test]
async fn test_ingest_subset() {
    common::log_init();
    let (mock, solr) = common::solr_server().await;
    Mock::given(method("POST"))
        .and(path("/solr/files/update/json/docs"))
        .respond_with(ResponseTemplate::new(200))
        .mount(&mock)
        .await;
    Mock::given(method("POST"))
        .and(path("/solr/latest/update/json/docs"))
        .respond_with(ResponseTemplate::new(200))
        .mount(&mock)
        .await;

    let config_dir = Utf8PathBuf::from(common::REPO_ROOT).join(".docker");
    let data_dir = config_dir.join("data/cmip5/cmip5/output2/INM");
    let drs_config = common::test_config();

    let result = freva::drs::ingest(
        &solr,
        &drs_config,
        &Some(data_dir.as_std_path()),
        10,
        None,
        None,
    )
    .await;
    assert!(result.is_ok());
    let report = result.unwrap();

    assert_eq!(
        1,
        report.datasets.len(),
        "unexpected number of dataset reports"
    );

    let cmip5_results = &report.datasets[0];
    assert_eq!(4, cmip5_results.sent);
}
